#include "2include.h"
#include <stdio.h>
#include <strings.h>

/*
NOTE:NOTE:NOTE:  
This version of the FGAPcovar program accomodates constraints
  on the "nu" parameters and so a bit of code must be checked or
  altered for specific constraints that may be desired.

  BEFORE USING THIS CODE BE SURE  THAT:
  1) The code following the lines  with: 
     a) "JH FILL pars from ogpars"
     b) "JH FILL pars from ogpars<-->ppars  in mkfindiv"
         have been correctly programmed.
      See "PARS" below.
  2) The function  mkuih is correct for the constraints
         that are desired. Make sure that the uih[2],....uih[npars-1],
         the score expressions associated with the constrained nu's,
         are correct.  See "JH-UIH" below.
  3) Change "JH MODIFY PRINTING OF PARAMETERS" as appropriate.
  4) Initialize cch so that the model is dom, recessive, or additive
         as appropriate.
  5) Initialize cchu1.
*/

float **matrix();
float *vector();
float phg();
float phgz();
float loglik();

int **imatrix();
long **lmatrix();
long *lvector();

void makeGT();
void gx();
void fpi();
void mkfindiv();
void inversse();
float mltmm();
float mkmuijg();
float mkmuu1ijg();
void mkuih();

void powell();

void free_matrix();
void free_imatrix();
void free_lmatrix();
void free_vector();
void free_lvector();

FILE *f1, *f2, *f3, *fweights;
long i, i1, i2, i3, itmp1, itmp2, itmp3, famcnt, ninfamcnt, k, j;
long k1, k2, k3, k4;
long jj1, jj2, jj3, jj4;
long ifam;
int nnp, nnpp;
int ffg, ffgg, mmg, mmgg;
float scoreUb = 0;


//utilg holds values for g0, g1, g2 for the efficient calculation
//   of d_g having all possible 9 2x2 submatrix's for the Variance.   
//  The appropriate row number of utilg is given by
//  3*fg + mg + 1;
int utilg[10][4] = { {0, 0, 0, 0}, {0, 2, 0, 0}, {0, 1, 1, 0}, {0, 1, 0, 1},
{0, 1, 1, 0}, {0, 0, 2, 0}, {0, 0, 1, 1},
{0, 1, 0, 1}, {0, 0, 1, 1}, {0, 0, 0, 2}
};

int rdx = 3;
int maxfamilysize = 10;
//int maxfamilysize = 4;
int maxNumberOfFamilies = 10000;
int *numberInEachFamily;

//rdxmaxfam will hold the number of genotypes possilbe for
//  a family of size maxfamilysize.
long rdxmaxfam;

//ncov is the number of covariates
//int ncov = 2;
int ncov = 2;


/** npars is the number of parameters.   
  In the basic case there will be 2*ncov "nu's"  plus 
   1 for frequency of parental genotype 1 , "gammma[1]" and
   1 for frequency of parental genotype 2 , "gammma[2]".
   Note: pars will always hold 2*ncov+2 parameters, but some
      of them may be functions of others.
      The "independent" parameters will be held and manipulated:
      particularly during the maximization using powell in
      a separate vector, ogpars.  How and where the values in
      pars (always of length 2*ncov+2) and those ogpars (always of
      length npars) are used in the program below will depend
      on what is most convenient and flexible. 
      see also "PARS" below.
*/

int npars = 2;

//unpgeno  is a utility integer used by gx().  It is declared global
//  so that it can be used to store the number of parental genotypes
//  consistent with  the children's genotypes that is calculated in gx.
int unpgeno;


/**cch: Its specification varies depending on whether the model is
     dominant, recessive, or additive.  For now during developement, 
     I will assume dominant. 
   Additive       cch={0,1,2}
   Dominant       cch={0,1,1}
   Recessiv       cch={0,0,1}
  */
//int cch[4] = { 0, 0, 1 };
int cch[4] = { 0, 1, 1 };
int cchu1[4] = { 0, 1, 1 };

//upgenos is a utility matrix used by gx() to store the 
//parental genotypes consistent with the childrens's genotypes.
int **upgenos;
int *ufamgeno;
int *ufamgeno1;

//on output of powell, *iter contains the number of iterations
//  taken
int itter;
int *iter = &itter;

float ffret;
float *fret = &ffret;


/** fpgiout is a utility matrix for use with fpi
   It will hold the parental genotypes consistant with a possible
   genotype given a possible parental genotype, It is used to store
   values for the inner (over possilbe parental genotypes of the
   quantity for which the expectation is being calculated  summation
   in the "Hu-Whittemore form".
  */
float **fpgiout;
/** fpgiout0 is a utility matrix for use with fpi
   It will hold the parental genotypes consistant with a familie's
   actual known genotypes. It is used to store values for the outer
   summation in the "Hu-Whittemore form".
  */
float **fpgiout0;

//nindiv is the number of individuals in a the dataset. 
//nfam   is the number of families in the dataset. 

long nindiv;
long nfam;

int iutil1[20];

/**  PARS
 pars contains the 2*ncov+2 parameters that are needed. 
  The parameter tested need be evaluated at only at  0, so 
  it is not included here.
  The first 2*ncov entries in pars contain a)the ncov covariate
  parameters for individual genotype=1and b)the ncov covariate parameters
  for individual genotype=2.  The 2*ncov+1st entry is "gammma[0]" (see below).
  The 2*ncov+2nd entry is "gammma[1]" (Two gammma parameters allow for
  non-Hardy Weinberg.  In the case of Hardy Weinberg, only one parameter
  gammma is necessary, and it may in that case be useful to have gammma be
  the population frequency of the deleterious allele).  
  entry is the parameter that will be tested.  In the case of HW, there
  will be only 2*ncov+1 parameters. 

      gammma[0] is the population frequency of genotype (1,1)
      gammma[1]  is the population frequency of genotype (1,2)
      gammma[2]  is the population frequency of genotype (2,2)
      by convention allele type 2 is taken as the uncommon
      deleterious allele.
      In this program gammma[1],  gammma[2] will be taken as the
      parameters that are to be estimated and gammma[0] will always
      be set equal to 1-gammma[1]-gammma[2].
      When not specified in the program code (for now they may not
         be given as part of the input) they will be estimated.
  */

float *pars;
float *ogpars;
float gammma[3];

/** like rho0,  the deltas must be specified by the
    user.  See below the lines after the space pointed to by 
    deltas is allocated.  For now with ncov = 2, I have
    put in some completely ficticious values for deltas.
   */
float deltas[101];


/** GT that holds all possible genotypes  that a family
 *with maxfamilysize members can have.
 */
long **GT;

/**  BB is a utility vector. It has been created particulary for
  *use in makeGT which is executed near the beginning of
  *the main procedure.
  */
int *BB;
//cntb is a utility variable used in makeGT.
int cntb = 0;

/** powrdx holds powers of rdx (usually rdx = 3). So that
 *computation of 3^(nc or nc2 or nc+1, etc) can be done only once.
 */
long *powrdx;

/**  phgdat phg:
 *There are 27 possibilities  (when Mother and Father are considered
 *distinct) for P(h|g) where g is the parental genotype and h is
 *the genotype for one child, since in the formulation in the
 *paper there are three genotypes g=(g_m, g_f) and each of
 *g_m, g_f, and h may take any of the values 0, 1, and 2.
 *depending on whether they are carrying 0, 1, or 2 mutations at the
 *locus of interest.
 *
 *In the coding directly below. the value x_ will be set (as appropriate)
 *to 9*g_f + 3*g_m + h  so  x_ is in [0,1,2,.....,26].  AND for this
 *program I will compute phg=Prob(h|g) by assuming that the phenocopy is
 *0 and that each parent hands down to a child one of the two alleles at
 *the locus of interest with probability .5.
 */
float phgdat[27] = { 1, 0, 0, .5, .5, 0, 0, 1, 0, .5,
    .5, 0, .25, .5, .25, 0, .5, .5, 0, 1,
    0, 0, .5, .5, 0, 0, 1
};
float *phgd1 = phgdat - 1;


//rho0 is the logit of the "phenocopy" rate.
//  rho0 rate = log((phenocopy rate)/(1-(phenocopy rate))) 
float rho0 = -2.751535;
//float rho0 = -2.751535;
//float rho0 = -10.0   0 phenocopy rate.;
//float rho0 = 2.751535; 
float tmp, tmp0, tmp1, tmp2, tmp3, tmp4, tmphh;

//uih will hold the score statistic for family i
float *uih;
//inneruihuih is used to accumulate uih while in the "inner loop" 
float *inneruih;

float **JJ;
float **subJ;
float **subJI;
float **Jbvg;
float **JbvgT;
float **Jutil3;
float **Jutil4;


/** famid is  number of family as it appears in
      the data.
    ninfam is number of family members.
    npgeno  is number of possible parental genotypes.
     *famgeno is the entire family genotype 
         (coded with 0, 1, and 2. -1 means missing.)
    **pgenos is matrix holding possible parental 
         genotypes given children's genotypes.
     *phenodat holds phenotype data for family.
    **fcovdat holds covariate data for family.
        including placeholders with missing covars for
        the parents.
    **findiv  holds utility floating point info associated with 
              each individual.
                1)  1.0
                2)  exp(sum((pars:h=1) * z)).   
                3)  exp(sum((pars:h=2) * z)).   
                4)   w_{ij} = q(z_{ij})^{y_{ij}} * [-p(z_{ij}]^{1-y_ij}}
    **iindiv  holds utility integer info associated with 
              each individual.

    fcovdat, famgeno, phenodat are all by 
          row order:
                father 
                mother 
                children in any order.
  */
struct famdat {
    long famid;
    int ninfam;
    int npgeno;
    int *famgeno;
    int **pgenos;
    int *phenodat;
    float **fcovdat;
    float **findiv;
    int **iindiv;
};

struct famdat *fdp;
struct famdat *ffdp;

char s1[150], s2[150];
char *s3;

//xii is a utility matrix used by powell()
float **xii;

/** Next four vector/matrix  are used by gx to efficiently  make
   the determination of possible parental genotypes given
   the known family genotype.

   parentgh gives, for each child genotype,
      the row indices in gpgenos of possible parental genotypes.
   parentg gives for a mother's known genotype 
      the row indices in gpgenos of possible parental genotypes.
   posparent is a utility matrix which keeps track for a given
      family genotype of the possible parental genotypes during
      the determination of the possible parental genotypes.
  */
int parentgh[3][10] = {
    {0, 1, 1, 0, 1, 1, 0, 0, 0, 0},
    {0, 0, 1, 1, 1, 1, 1, 1, 1, 0},
    {0, 0, 0, 0, 0, 1, 1, 0, 1, 1}
};
int parentg[3][10] = {
    {0, 1, 0, 0, 1, 0, 0, 1, 0, 0},
    {0, 0, 1, 0, 0, 1, 0, 0, 1, 0},
    {0, 0, 0, 1, 0, 0, 1, 0, 0, 1}
};
int posparent[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

int gpgenos[10][2] = {
    {0, 0},
    {0, 0},
    {0, 1},
    {0, 2},
    {1, 0},
    {1, 1},
    {1, 2},
    {2, 0},
    {2, 1},
    {2, 2}
};
int ckmkuih;
long whichfam;
float waldstat;
int wsdf;

//smodeltype is an indicator of the type of influence the user
//   specifices that genotype has on disease status.
int smodeltype = -1;

//cmodeltype is an indicator of the type of influence the user
//   specifices that genotype has on covariates.  This
//   is equivalent to corresponding types of constraints on the
//   model parameters, nu_1  and nu_2. 
int cmodeltype = -1;

main()
{
// f3 = fopen("outsimstmp","a");
    char modedesc[5][20]={{"None"},{"Recessive"},{"Dominant"},{"Additive"}};
    fprintf(stderr, "\nEnter the number of covariates(Including the \n");
    fprintf(stderr, "           Intercept if there is one)-->");
    scanf("%ld", &ncov);
    fprintf(stderr, "            %ld\n", ncov);

   npars = 2 + 2*ncov; 
/* The default number of parameters is 2 * (number of covariatesIncludingIntercept) + 2.
     The extra +2 are for the prevelance of genotypes 1 and 2.
   If the constraints, nu_1 = 0, nu_2 arbitrary, or nu_1=nu_2=arbitrary  are selected
     by the user than npars will be set at 2 + ncov.
*/
   
/*
    fprintf(stderr, "\nEnter the number of parameters-->");
    scanf("%ld", &npars);
    fprintf(stderr, "%ld\n", npars);
*/

    while ((smodeltype < 1) || (smodeltype > 3)) {
        smodeltype = 2;
// default smodeltype is Dominant.
        fprintf(stderr,
                "\nEnter a model_type number for effect of genotype on status\n");
        fprintf(stderr,
                "           Recessive(1), Dominant(2), or Additive(3)-->");
        scanf("%ld", &smodeltype);
        if ((smodeltype < 1) || (smodeltype > 3)) {
            printf("\nPlease enter a model type 1, 2, or 3");
        }
    }
        fprintf(stderr,"            %ld\n", smodeltype);

    if(ncov != 0){
    while ((cmodeltype < 0) || (cmodeltype > 3)) {
        cmodeltype = 0;
// default cmodeltype is "no-effect".
        fprintf(stderr,
                "\nEnter a model_type number for effect of genotype on covariates.\n");
        fprintf(stderr,"For further explanation of this, see the documentation.\n");
        fprintf(stderr,
                "           None(0), Recessive(1), Dominant(2), or Additive(3)-->");
        scanf("%ld", &cmodeltype);
        if ((cmodeltype < 0) || (cmodeltype > 3)) {
            printf("\nPlease enter a model type 0, 1, 2, or 3");
        }
    }
        fprintf(stderr,"            %ld\n", cmodeltype);
  }
  else { cmodeltype = 0;}

    if (smodeltype == 1) {
        cchu1[0] = 0;
        cchu1[1] = 0;
        cchu1[2] = 1;
    } else if (smodeltype == 2) {
        cchu1[0] = 0;
        cchu1[1] = 1;
        cchu1[2] = 1;
    } else if (smodeltype == 3) {
        cchu1[0] = 0;
        cchu1[1] = 1;
        cchu1[2] = 2;
    }

    if (cmodeltype == 1) {
        cch[0] = 0;
        cch[1] = 0;
        cch[2] = 1;
        npars = 2 + ncov;
    } else if (cmodeltype == 2) {
        cch[0] = 0;
        cch[1] = 1;
        cch[2] = 1;
        npars = 2 + ncov;
    } else if (cmodeltype == 3) {
        cch[0] = 0;
        cch[1] = 1;
        cch[2] = 2;
        npars = 2 + ncov;
    }

/* Neither rho0 nor delta(s) are used in this version of the program
   in which there is a file, wweights, giving the weights=Prob(y=1|z) that
   are to be used.

 fprintf(stderr,"Enter the phenocopy rate-->");
 scanf("%lf",&rho0);
 fprintf(stderr,"%lf\n",rho0);
 rho0=log(rho0/(1.0-rho0));

    for (i = 0; i < 101; i++) {
        deltas[i] = 0;
    }

for(i = 0; i < ncov; i++){
 fprintf(stderr,"Enter the deltas[%ld]-->",(i+1));
 scanf("%lf",&deltas[i+1]);
 fprintf(stderr,"%lf\n",deltas[i+1]);
}
*/

/* First determine how many individuals there are,
    how many families there are,  and how many
    individuals there are in each family.
*/

    numberInEachFamily = ivector(1, maxNumberOfFamilies);

    nindiv = 0;
    nfam = 0;
    itmp1 = -997;
    itmp2 = itmp1;
    ninfamcnt = 0;
    f1 = fopen("wpedigree.dat", "r");
    while (fgets(s1, 200, f1) != NULL) {
        sscanf(s1, "%ld", &itmp2);
        ninfamcnt++;
        if (itmp2 != itmp1) {
            nfam++;
            if (nfam > 1) {
                numberInEachFamily[nfam - 1] = ninfamcnt - 1;
                ninfamcnt = 1;
            }
            itmp1 = itmp2;
        }
        nindiv++;
    }
    fclose(f1);
    numberInEachFamily[nfam] = ninfamcnt;


//I will waste familyData[0]; so that
//  useful index of familyData is from 1,...,nfam.
    struct famdat familyData[nfam + 1];
    ffdp = familyData;


    for (i = 1; i <= nfam; i++) {
        fdp = &familyData[i];
        (*fdp).ninfam = numberInEachFamily[i];
        familyData[i].ninfam = numberInEachFamily[i];
        (*fdp).famgeno = ivector(1, numberInEachFamily[i]);
        (*fdp).fcovdat = matrix(1, numberInEachFamily[i], 1, ncov);
        (*fdp).findiv = matrix(1, numberInEachFamily[i], 1, 10);
        (*fdp).phenodat = ivector(1, numberInEachFamily[i]);
        (*fdp).pgenos = imatrix(1, 9, 1, 2);
    }

    free_ivector(numberInEachFamily, 1, maxNumberOfFamilies);


    pars = vector(1, 2 * ncov + 2);
    ogpars = vector(1, npars);
/*NOTE: At this point I am am writing the program.
  The analysis calls for I. the estimation of the parameters;
  and II. the use of them for the computation 
  the variance of the score.  Here I am setting the
  parameters without estimating them. After I have done 
  II, I will return and do I
*/



//for(i=1; i<= 2*ncov; i++){ pars[i] = .2 + sin(((double)i)/6.0);}
    for (i = 1; i <= 2 * ncov; i++) {
        pars[i] = 0;
    }
    pars[2 * ncov + 1] = 1.0 / 3.0;
    pars[2 * ncov + 2] = 1.0 / 3.0;

    for (i = 1; i <= npars; i++) {
        ogpars[i] = 0;
    }
    ogpars[npars - 1] = 1.0 / 3.0;
    ogpars[npars] = 1.0 / 3.0;

    gammma[0] = 1.0 - ogpars[npars - 1] - ogpars[npars];
    gammma[1] = ogpars[npars - 1];
    gammma[2] = ogpars[npars];

    powrdx = lvector(1, maxfamilysize);
    powrdx[1] = rdx;
    for (i1 = 2; i1 <= maxfamilysize; i1++) {
        powrdx[i1] = rdx * powrdx[i1 - 1];
    }

    rdxmaxfam = (long) pow((double) rdx, (double) maxfamilysize);
    GT = lmatrix(1, rdxmaxfam, 1, maxfamilysize);
    for (i1 = 1; i1 <= rdxmaxfam; i1++) {
        for (i2 = 1; i2 <= maxfamilysize; i2++) {
            GT[i1][i2] = 0;
        }
    }

    BB = ivector(1, maxfamilysize);
    for (i = 1; i <= maxfamilysize; i++) {
        BB[i] = 0;
    }

    fpgiout0 = matrix(1, 27, 1, 3);
    fpgiout = matrix(1, 27, 1, 3);

    upgenos = imatrix(1, 9, 1, 3);
    ufamgeno = ivector(1, maxfamilysize);
    ufamgeno1 = ivector(1, maxfamilysize);

    xii = matrix(1, npars, 1, npars);
    for (i = 1; i <= npars; i++) {
        for (i1 = 1; i1 <= npars; i1++) {
            xii[i][i1] = 0;
        }
        xii[i][i] = 1.0;
    }



    uih = vector(1, npars + 1);
    inneruih = vector(1, npars + 1);
    JJ = matrix(1, npars + 1, 1, npars + 1);
    subJ = matrix(1, npars, 1, npars);
    subJI = matrix(1, npars, 1, npars);
    Jbvg = matrix(1, npars, 1, 1);
    JbvgT = matrix(1, 1, 1, npars);
    Jutil3 = matrix(1, 1, 1, npars);
    Jutil4 = matrix(1, 1, 1, 1);

    itmp1 = 0;
    itmp2 = -979;               //this keeps the number of the current family.
    famcnt = 0;
    ninfamcnt = 0;
    fdp = &familyData[1];
    f1 = fopen("wpedigree.dat", "r");
    if( ncov > 0){
    f2 = fopen("wcovars.dat", "r");
    }
    fweights = fopen("wweights.dat", "r");
    itmp3 = 0;
// During the reading of the data and immediately afterwards, itmp3
//   is a flag that is set to 1 if an error is detected.    

    while (fgets(s1, 200, f1) != NULL) {
    if( ncov > 0){
        fgets(s2, 200, f2);
     }

        itmp1++;
        ninfamcnt++;
        sscanf(s1, "%ld %ld %ld %ld %ld %ld %ld %ld %ld", &iutil1[1],
               &iutil1[2], &iutil1[3], &iutil1[4], &iutil1[5], &iutil1[6],
               &iutil1[7], &iutil1[8], &iutil1[9]);

//Check to see if data is for a new family
//  NOTE: IT IS a new family on the first loop for family 1.
        if (iutil1[1] != itmp2) {
            itmp2 = iutil1[1];
            famcnt++;
            fdp = &familyData[famcnt];
            if (famcnt > 1) {
                ninfamcnt = 1;
                (*fdp).famid = iutil1[1];
                if ((iutil1[8] * iutil1[9]) == 0) {
                    (*fdp).famgeno[ninfamcnt] = -1;
                } else {
                    (*fdp).famgeno[ninfamcnt] = iutil1[8] + iutil1[9] - 2;
                }

            }                   //END if(famcnt > 1)
        }                       //END if(iutil1[1] != itmp2)

        if ((iutil1[6] < 0) || (iutil1[6] > 2)) {
            printf
                ("\nERROR:  Family %ld: This family has an impermissible penotype.\n",
                 iutil1[1]);
            printf
                ("Each individual must have a phenotype of 0(missing) or 1 or 2\n");
            itmp3 = 1;
        }

        if ((iutil1[8] * iutil1[9]) == 0) {
            (*fdp).famgeno[ninfamcnt] = -1;
        } else {
            (*fdp).famgeno[ninfamcnt] = iutil1[8] + iutil1[9] - 2;
        }

        if ((iutil1[8] * iutil1[9] == 0)
            && ((iutil1[8] != 0) | iutil1[9] != 0)) {
            printf
                ("\nERROR: Family %ld: An individual has only one of two alleles known.\n",
                 iutil1[1]);
            itmp3 = 1;
        }
        if ((iutil1[8] * iutil1[9] == 0)
            && ((iutil1[3] != 0) | iutil1[4] != 0)) {
            printf
                ("\nERROR:  Family %ld: This child  has an unknown  genotype.\n",
                 iutil1[1]);
            itmp3 = 1;
        }

        (*fdp).phenodat[ninfamcnt] = iutil1[6];
        (*fdp).famid = itmp2;
//        if ((iutil1[3] == 0) && (iutil1[4] == 0) && (iutil1[5] == 1)) {
        if ((iutil1[3] == 0) && (iutil1[4] == 0)) {
            if ((iutil1[8] == 0) && (iutil1[9] == 0)) {
                (*fdp).famgeno[ninfamcnt] = -1;
            } else {
                (*fdp).famgeno[ninfamcnt] = iutil1[8] + iutil1[9] - 2;
            }
        }

        s3 = strtok(s2, " ");
        for (i = 1; i <= ncov; i++) {
            sscanf(s3, "%lf", &tmp1);
            (*fdp).fcovdat[ninfamcnt][i] = tmp1;
            s3 = strtok(NULL, " ,\t");
        }
        fscanf(fweights, "%lf", &tmp1);
        (*fdp).findiv[ninfamcnt][4] = tmp1;
    }                           //END reading pedigree and covariance data. 
    fclose(f1);
    if( ncov > 0){
    fclose(f2);
     }
    fclose(fweights);

    for (i = 1; i <= famcnt; i++) {
        fdp = &familyData[i];
        gx((*fdp).famgeno, (*fdp).ninfam, (*fdp).famid);
        if (unpgeno == 0) {
            printf
                ("\nERROR: Family %ld: Family genotypes are inconsistant.\n",
                 familyData[i].famid);
            itmp3 = 1;
        }
    }

    for (i = 1; i <= famcnt; i++) {
        if ((familyData[i].famgeno[1] != -1) &&
            (familyData[i].famgeno[2] == -1)) {
            itmp3 = 1;
            printf
                ("\nERROR: Family %ld: Mother's genotype is unknown.  Father's is known.\n",
                 familyData[i].famid);
            printf
                (" If only one parent has known genotype, it must be the mother.\n");
            printf
                (" Change the data, \(interchange the lines for mother and father in\n");
            printf
                (" the wpedigree.dat, wcovars.dat, and wweights.dat file\051  for mother\n");
            printf
                (" to that of father's \(and remember to assign 'father' to  sex=2=female\n");
            printf(" and 'mother' to sex=1\051.\n");
            printf(" Then rerun.\n");
        }
    }
    if (itmp3 == 1) {
        exit(1);
    }
//initialize pars
    for (i = 1; i <= 2 * ncov + 2; i++) {
        pars[i] = .15 + i / 200.0;
    }
    for (i = 1; i <= npars; i++) {
        ogpars[i] = .15 + i / 200.0;
    }

//NOW having read the data and knowing the certain parameters for each family
//  it is possible to use what is known to fill in other familiy
//  features that depend on knowing these parameters.

/*
ckmkuih = 1; (*fdp)=familyData[4];
printf("TOHERE TESTING gx %ld\n", (*fdp).famid);
for(i1 = 1; i1 <= 8; i1++){
ufamgeno[i1] = 0;
}
ufamgeno[1] = -1;
gx((*fdp).famgeno,(*fdp).ninfam,11);
ckmkuih = 0; 
for(itmp1 = 1; itmp1 <=unpgeno; itmp1++){
printf("TOHERE gxtst  %ld  %ld\n",upgenos[itmp1][1],upgenos[itmp1][2]);
}
exit(1);
*/


    for (i = 1; i <= nfam; i++) {
        fdp = &familyData[i];
        for (j = 1; j <= (*fdp).ninfam; j++) {
            (*fdp).findiv[j][1] = 1.0;
            mkfindiv(ogpars, fdp);
        }


        gx((*fdp).famgeno, (*fdp).ninfam, (*fdp).famid);
        (*fdp).npgeno = unpgeno;
        for (i1 = 1; i1 <= (*fdp).npgeno; i1++) {
            (*fdp).pgenos[i1][1] = upgenos[i1][1];
            (*fdp).pgenos[i1][2] = upgenos[i1][2];
        }
    }                           //END for(i = 1; i <= nfam; i++)

    makeGT(BB, 0);

    powell(ogpars, xii, npars, .000001, iter, fret, loglik);

    tmp1 = (atan(ogpars[npars - 1]) + 3.14159 / 2.0) / 3.14159;
    tmp2 = 1 - tmp1;
    tmp2 = tmp2 * (atan(ogpars[npars]) + 3.14159 / 2.0) / 3.14159;

    gammma[0] = 1.0 - tmp1 - tmp2;
    tmp0 = gammma[0];
    gammma[1] = tmp1;
    gammma[2] = tmp2;

//NOTE CHANGE next clause to fill pars  from the now
//   established estimates of the parameters stored in
//   ogpars.
//JH FILL pars from ogpars

if(cmodeltype == 0){
//UNCONSTRAINED
     for(i = 1; i <= npars-2; i++){
       pars[i] = ogpars[i];
     } 
}
else if(cmodeltype == 1){
//RECESSIVE  covariates.
     for(i = 1; i <= npars-2; i++){
       pars[i] = 0;
       pars[i + (npars-2)] = ogpars[i];
     } 
}
else if(cmodeltype == 2){
//DOMINANT  covariates.
     for(i = 1; i <= npars-2; i++){
       pars[i] = ogpars[i];
       pars[i + (npars-2)] = ogpars[i];
     } 
}
else if(cmodeltype == 3){
//ADDITIVE  covariates.
     for(i = 1; i <= npars-2; i++){
       pars[i] = ogpars[i];
       pars[i + (npars-2)] = 2 * ogpars[i];
     } 
}

    pars[2 * ncov + 2] = ogpars[npars];
    pars[2 * ncov + 1] = ogpars[npars - 1];


//This next use of mkfindiv is not strictly necessary, since the
// likelihood function used in powell calls mkfindiv for each
//  new set of values of the parameters.  However, placed here, it
//  allows me to set arbitrary values of the parameters after powell
//  for testing purposes.
    for (i = 1; i <= nfam; i++) {
        fdp = &familyData[i];
        for (j = 1; j <= (*fdp).ninfam; j++) {
            (*fdp).findiv[j][1] = 1.0;
            mkfindiv(ogpars, fdp);
        }
    }


/*
printf("\n");
for(i1 = 1; i1 <= rdxmaxfam; i1++){
for(i2 = 1; i2 <= maxfamilysize; i2++){
printf("%2ld",GT[i1][i2]);
}
printf("\n");
}
printf("\n");
printf("%lf \n",loglik(pars));
*/
/*===============================================================*/
/*

for(i1 = 1; i1 <= nfam; i1++){
fdp = &ffdp[i1];
printf("\n\n i1=%ld famid=%ld\n", i1, (*fdp).famid); 
printf("ninfam=%ld\n", (*fdp).ninfam); 
printf("npgeno=%ld\n", (*fdp).npgeno); 
printf("   famgeno\n"); 
for(i2 = 1; i2 <= (*fdp).ninfam; i2++){
printf("%ld  ", (*fdp).famgeno[i2]); 
}
printf("\n"); 

printf("   pgenos\n"); 
for(i2 = 1; i2 <= (*fdp).npgeno; i2++){
printf(" %ld   %ld\n",(*fdp).pgenos[i2][1], (*fdp).pgenos[i2][2]);
}

printf("   phenodat\n"); 
for(i2 = 1; i2 <= (*fdp).ninfam; i2++){
printf("%ld  ", (*fdp).phenodat[i2]); 
}

printf("\n"); 
printf("   fcovdat\n"); 
for(i2 = 1; i2 <= (*fdp).ninfam; i2++){
for(i3 = 1; i3 <= ncov; i3++){

printf( "%lf ",(*fdp).fcovdat[i2][i3]);
}
printf("\n"); 
}

}

//END OF debug printing of family data saved in familyDat
*/
/*=======================================================================*/

    for (i1 = 1; i1 <= npars + 1; i1++) {
        for (i2 = 1; i2 <= npars + 1; i2++) {
            JJ[i1][i2] = 0;
        }
    }
//BEGIN LOOP OVER FAMILIES
    for (ifam = 1; ifam <= nfam; ifam++) {
        fdp = &ffdp[ifam];

/*
if((*fdp).famid == 19){
printf("TOHEREJJBdiff  %lf\n",JJ[1][1]);
}
*/

        ckmkuih = 1;
        fpi(fdp, (*fdp).famgeno, (*fdp).pgenos, (*fdp).npgeno, fpgiout0);
/*
if((*fdp).famid == 19){
printf("TOHERE wpp %lf %lf\n", fpgiout0[1][3], fpgiout0[2][3]);
}
*/
        for (nnp = 1; nnp <= (*fdp).npgeno; nnp++) {
            ffg = fpgiout0[nnp][1];
            mmg = fpgiout0[nnp][2];

            ckmkuih = 1;

//Now a slight "diversion" to accrue the score.
            for (i1 = 1; i1 <= (*fdp).ninfam; i1++) {
                ufamgeno[i1] = (*fdp).famgeno[i1];
            }
            for (i1 = 1; i1 <= (*fdp).ninfam; i1++) {
                ufamgeno[1] = ffg;
                ufamgeno[2] = mmg;
            }
            mkuih(fdp, ufamgeno);
            scoreUb += fpgiout0[nnp][3] * uih[1];
//printf("TOHERE uih nnp=%ld famid=%ld wp=%6.3lf fg=%ld score=%6.3lf uih=%6.3lf\n",
//                  nnp, (*fdp).famid,fpgiout0[nnp][3],ffg,scoreUb,uih[1]);
            if (ckmkuih == 1) {
/*
if((*fdp).famid == 19){
printf("TOHERE uih nnp=%ld famid=%ld wp=%6.3lf fg=%ld score=%6.3lf uih=%6.3lf\n\n",
                  nnp, (*fdp).famid,fpgiout0[nnp][3],ffg,scoreUb,uih[1]);
}
*/
            }
//END of "diversion"
            ckmkuih = 0;


/**k1 and k2 are the lower and upper bound on the row numbers of
   GT that need to be examined. These are the rows that have the
   father and mother genotypes ffg and mmg.  It's possilbe that
   something more efficient that would skip impossible genotypes
   that are rows in GT
  */
            k1 = powrdx[(*fdp).ninfam - 1] * ffg + powrdx[(*fdp).ninfam -
                                                          2] * mmg + 1;
            k2 = powrdx[(*fdp).ninfam - 1] * ffg + powrdx[(*fdp).ninfam -
                                                          2] * (mmg + 1);
/*
            if ((*fdp).famid == 8) {
                printf("TOHEREk1k2 k1=%ld k2=%ld\n", k1, k2);
            }
*/
            itmp1 = maxfamilysize - (*fdp).ninfam;
            for (k3 = k1; k3 <= k2; k3++) {
                for (i1 = 1; i1 <= (*fdp).ninfam; i1++) {
                    ufamgeno1[i1] = GT[k3][itmp1 + i1];
                }

                tmphh = 1;
                for (i1 = 3; i1 <= (*fdp).ninfam; i1++) {
                    tmphh *=
                        phgz(ufamgeno1[1], ufamgeno1[2], ufamgeno1[i1],
                             (*fdp).findiv[i1][2], (*fdp).findiv[i1][3]);
                }
                if (tmphh == 0) {
                    continue;
                }

                if ((*fdp).famgeno[1] == -1) {
                    ufamgeno1[1] = -1;
                }
                if ((*fdp).famgeno[2] == -1) {
                    ufamgeno1[2] = -1;
                }
                gx(ufamgeno1, (*fdp).ninfam, (*fdp).famid);
                fpi(fdp, ufamgeno1, upgenos, unpgeno, fpgiout);


                for (jj1 = 1; jj1 <= npars + 1; jj1++) {
                    inneruih[jj1] = 0;
                }
/*
if((*fdp).famid == 19){
printf("TOHEREinnB  BEGIN INNERLOOP\n");
}
*/
                for (nnpp = 1; nnpp <= unpgeno; nnpp++) {
                    ufamgeno1[1] = upgenos[nnpp][1];
                    ufamgeno1[2] = upgenos[nnpp][2];

                    mkuih(fdp, ufamgeno1);
                    for (jj1 = 1; jj1 <= npars + 1; jj1++) {
                        inneruih[jj1] += uih[jj1] * fpgiout[nnpp][3];
/*
if((*fdp).famid == 19 && jj1 == 1){
printf("TOHEREjiner uih=%lf wght=%lf tmphh=%lf\n",uih[jj1],
            fpgiout[nnpp][3], tmphh);
}
*/
                    }
                }               //END inner nnpp LOOP
/*
if((*fdp).famid == 19){
printf("TOHEREinnE  END INNERLOOP\n");
}
*/
                for (jj1 = 1; jj1 <= npars + 1; jj1++) {
                    for (jj2 = jj1; jj2 <= npars + 1; jj2++) {
//NOTE: Here is where the JJ matrix is actually being accrued
                        JJ[jj1][jj2] +=
                            inneruih[jj1] * inneruih[jj2] * tmphh *
                            fpgiout0[nnp][3];
/*
if((*fdp).famid == 19 && jj1==1 && jj2 == 1){
printf("TOHEREfadd inn=%lf tmphh=%lf fpg0=%lf JJcont=%lf\n",inneruih[jj1],
                                tmphh,fpgiout0[nnp][3],
                            inneruih[jj1] * inneruih[jj2] * tmphh *
                            fpgiout0[nnp][3]);
} 
*/
/*
                        if (tmphh > 0 && jj1 == 1 && jj2 == 1) {
                            if ((*fdp).famid == 19) {
                                printf
                                    ("TOHEREJJ famid=%ld  k3=%ld nnp=%ld nnpp=%ld j1=%ld j2=%ld   u1=%lf              u2=%lf  JJ=%lf         fpin=%lf        ingpp=%lf        ingmm=%lf fpout=%lf tmphh=%lf\n",
                                     (*fdp).famid, k3, nnp, nnpp, jj1, jj2,
                                     uih[jj1], uih[jj2] * uih[jj2], JJ[1][1],
                                     fpgiout[nnpp][3], fpgiout[nnpp][1],
                                     fpgiout[nnpp][2], fpgiout0[nnp][3],
                                     tmphh);
                            }
                        }
*/
                    }
                }

            }                   //END loop over GT for ffg,mmg

        }                       //END outer nnp LOOP
/*
if((*fdp).famid == 19){
printf("TOHEREJJEdiff  %lf\n",JJ[1][1]);
}
*/
    }                           //END ifam loop LOOP OVER FAMILIES
    for (jj1 = 1; jj1 <= npars + 1; jj1++) {
        for (jj2 = jj1; jj2 <= npars + 1; jj2++) {
            JJ[jj2][jj1] = JJ[jj1][jj2];
        }
    }

/*
    printf("\n JJ\n");
    for (jj1 = 1; jj1 <= 2 * ncov + 3; jj1++) {
        for (jj2 = 1; jj2 <= 2 * ncov + 3; jj2++) {
            printf("%7.3lf ", JJ[jj1][jj2]);
        }
        printf("\n");
    }

    printf("\n");
*/


    for (jj1 = 1; jj1 <= npars; jj1++) {
        Jbvg[jj1][1] = JJ[1][jj1 + 1];
        JbvgT[1][jj1] = JJ[1][jj1 + 1];
        for (jj2 = 1; jj2 <= npars; jj2++) {
            subJ[jj1][jj2] = JJ[jj1 + 1][jj2 + 1];
        }
    }

/*
    printf("\n subJ\n");
    for (jj1 = 1; jj1 <= npars; jj1++) {
        for (jj2 = 1; jj2 <= npars; jj2++) {
            printf("%7.3lf ", subJ[jj1][jj2]);
        }
        printf("\n");
    }
    printf("\n");
*/

    inversse(subJ, npars, subJI);

    waldstat = 0;
    for (i1 = 1; i1 <= npars - 2; i1++) {
        tmp = 0;
        for (i2 = 1; i2 <= npars - 2; i2++) {
            tmp += ogpars[i2] * subJ[i1][i2];
        }
        waldstat += ogpars[i1] * tmp;
    }


    tmp = 0;
    for (i1 = 1; i1 <= npars - 2; i1++) {
        if (subJ[i1][i1] >= tmp) {
            tmp = subJ[i1][i1];
        }
    }

    wsdf = 0;
    for (i1 = 1; i1 <= npars - 2; i1++) {
//if(JJI[i1+1][i1+1] >= .000001 * tmp) { wsdf++;}
        if (subJ[i1 + 1][i1 + 1] >= .000001 * tmp) {
            wsdf++;
        }
    }

/*
    printf("\n subJI\n");
    for (jj1 = 1; jj1 <= 2 * ncov + 2; jj1++) {
        for (jj2 = 1; jj2 <= 2 * ncov + 2; jj2++) {
            printf("%9.5lf ", subJI[jj1][jj2]);
        }
        printf("\n");
    }
    printf("\n");
*/

    mltmm(JbvgT, 1, npars, subJI, npars, Jutil3);
    mltmm(Jutil3, 1, npars, Jbvg, 1, Jutil4);

/*
    printf("NFS=%f Num=%lf stdNFS=%lf  JJ11=%lf %lf\n\n",
           scoreUb / sqrt((JJ[1][1] - Jutil4[1][1])),
           scoreUb, sqrt((JJ[1][1] - Jutil4[1][1])), JJ[1][1], Jutil4[1][1]
              );
*/
    printf("\n\nTDT with Covariates: ANALYSIS\n");
    printf
        ("==============================================================\n\n");
    printf("Number of Families: %ld\n", nfam);
    printf("Number of Covariates: %ld\n", ncov);
    printf("Number of Parameters: %ld\n", npars);
    printf("Influence of genotype on disease status: %s\n",
                                            modedesc[smodeltype]);
    if(ncov != 0){
    printf("Influence of genotype on covariates:     %s\n\n\n",
                                            modedesc[cmodeltype]);
    }
    else{
    printf("\n\n");
    }
/*
    printf("gammas0,1,2   %6.4lf     %6.4lf (%7.5lf)     %6.4lf (%7.5lf)\n\n",
           gammma[0], gammma[1], sqrt(subJI[npars - 1][npars - 1]),
           gammma[2], sqrt(subJI[npars][npars]));
*/
    printf("Prevalence of genotype 0: %6.4lf\n",gammma[0]);
    printf("Prevalence of genotype 1: %6.4lf (%7.5lf)\n",
                       gammma[1],sqrt(subJI[npars - 1][npars - 1]));
    printf("Prevalence of genotype 2: %6.4lf (%7.5lf)\n",
                       gammma[2],sqrt(subJI[npars][npars]));
//JH MODIFY PRINTING OF PARAMETERS
    for (i1 = 1; i1 <= npars - 2; i1++) {
        printf("par-%ld  %10.5lf (%lf)\n",
               i1, ogpars[i1], sqrt(subJI[i1][i1]));
    }
    printf("\n");
    if(npars > 2){
       printf("maximizedLogLike=%.4lf  WaldStatNus=%.4lf  df=%ld\n\n",
           *fret, waldstat, wsdf);
    }
     else{
       printf("maximizedLogLike=%.4f\n\n",
           *fret);
    }
    printf("Non-Founders Statistic=%.4lf  NFSdenominator=%.5lf\n",
           scoreUb / sqrt((JJ[1][1] - Jutil4[1][1])),
           sqrt((JJ[1][1] - Jutil4[1][1]))
        );
    printf
        ("==============================================================\n\n");

    fflush(stdout);
//    fprintf(f3,"============================================\n");
/*
    for(i1 = 1; i1 <= npars-2; i1++){
    fprintf(f3,"%10.5lf\n", ogpars[i1]);
   }
    fprintf(f3,"%12.4lf\n",*fret);
    fprintf(f3,"%12.4lf\n", waldstat);
    fprintf(f3,"%ld\n", wsdf);
    fprintf(f3,"%lf\n", scoreUb / sqrt((JJ[1][1] - Jutil4[1][1])));
*/

    fflush(f3);
    fflush(stdout);

    free_lmatrix(GT, 1, rdxmaxfam, 1, maxfamilysize);
    free_vector(pars, 1, 2 * ncov + 2);
    free_vector(ogpars, 1,  npars);
    free_vector(uih, 1, npars + 1);
    free_vector(inneruih, 1, npars + 1);
    free_matrix(JJ, 1, npars + 1, 1, npars + 1);
    free_matrix(subJ, 1, npars, 1, npars);
    free_matrix(subJI, 1, npars, 1, npars);
    free_matrix(Jbvg, 1, npars, 1, 1);
    free_matrix(JbvgT, 1, 1, 1, npars);
    free_matrix(Jutil3, 1, 1, 1, npars);
    free_matrix(Jutil4, 1, 1, 1, 1);
    free_lvector(powrdx, (long) 1, (long) maxfamilysize);
    free_ivector(ufamgeno, 1, maxfamilysize);
    free_ivector(ufamgeno1, 1, maxfamilysize);
    free_ivector(BB, 1, maxfamilysize);
    free_imatrix(upgenos, 1, 9, 1, 3);
    free_matrix(fpgiout, 1, 27, 1, 3);
    free_matrix(fpgiout0, 1, 27, 1, 3);
    free_matrix(xii, 1, npars, 1, npars);
    fflush(NULL);
}                               //END main


/** Given the genotypes of the parents (one or both possibly unknown), 
     and the genotypes of the children (all are assumed known), this
     function determines a matrix, upgenos, of possible parental genotypes.
     If only one parental genotype is unknown, then the original data will
     be rearranged so that it is the mother whose genotype is known. 
     upgenos[i][1] is father.  upgenos[i][2] is mother. nninfamily is
     the number of children + 2.

     Genotypes are (0,0)=0, (1,0)/(0,1)=1, and (1,1)=2.  For a parent
       the genotype  can be missing, this is coded as -1.
  */

/** Given the genotypes of the parents (one or both possibly unknown), 
     and the genotypes of the children (all are assumed known), this
     function determines a matrix, upgenos, of possible parental genotypes.
     If only one parental genotype is unknown, then the original data will
     be rearranged so that it is the mother whose genotype is known. 
     upgenos[i][1] is father.  upgenos[i][2] is mother. nninfamily is
     the number of children + 2.   the number of possible genotypes is
     put into the global variable unpgeno.

     Genotypes are (0,0)=0, (1,0)/(0,1)=1, and (1,1)=2.  For a parent
       the genotype  can be missing, this is coded as -1.
  */
void gx(ufamg, nninfam, famid)
int nninfam;
int *ufamg;
long famid;
{
    int i, i1, i2;
    int ih;
    int hcount = 0;
    int childin[3] = { 0, 0, 0 };

    unpgeno = 0;
    for (i1 = 1; i1 <= 9; i1++) {
        posparent[i1] = 1;
    }
    if (ufamg[1] != -1) {
        unpgeno = 1;
        upgenos[1][1] = ufamg[1];
        upgenos[1][2] = ufamg[2];
    } else {
        for (i = 3; i <= nninfam; i++) {
            if (hcount == 3) {
                continue;
            }
            ih = ufamg[i];
            if (childin[ih] != 1) {
                for (i1 = 1; i1 <= 9; i1++) {
                    posparent[i1] *= parentgh[ih][i1];
//printf("TOHEREpos ih=%ld i=%ld i1=%ld pihi1=%ld\n",
//                            ih,i,i1,parentgh[ih][i1]);

/*
for(i3 = 1; i3 <= 9; i3++){
printf(" %ld", posparent[i3]);
}
printf("\n\n");
*/
                }               //end i1 loop
//printf("\n");
                childin[ih] = 1;
            }
        }
        if (ufamg[2] != -1) {
            for (i1 = 1; i1 <= 9; i1++) {
                posparent[i1] *= parentg[ufamg[2]][i1];
            }
        }
/*
printf("\n\nTOHEREFinalppsx ");
for(i3 = 1; i3 <= 9; i3++){
printf(" %ld", posparent[i3]);
}
printf("\n\n");
*/
        for (i1 = 1; i1 <= 9; i1++) {
            if (posparent[i1] == 1) {
//printf("TOHERE is 1? i1=%ld unp=%ld\n",i1,unpgeno);
                unpgeno++;
                upgenos[unpgeno][1] = gpgenos[i1][0];
                upgenos[unpgeno][2] = gpgenos[i1][1];
            }
        }
    }                           //END else

}                               //END gx

void makeGT(Gg, dp)
int *Gg;
int dp;
{
    int *BB;
    int i, j;
    int cp;
    BB = ivector(1, maxfamilysize);
    cp = dp + 1;

    for (j = 1; j <= maxfamilysize; j++) {
        BB[j] = Gg[j];
    }

    for (i = 0; i < rdx; i++) {
        BB[cp] = i;
        if (cp < maxfamilysize) {
            makeGT(BB, cp);
        } else {
            cntb++;
            for (j = 1; j <= maxfamilysize; j++) {
                GT[cntb][j] = BB[j];
            }
        }
    }
    free_ivector(BB, 1, maxfamilysize);
}                               //END makeGT

/**   phg:  given the father genotype, the mother genotype,
   and the child genotype, phg calculates
   the index of the appropriate entry in phg 
   and returns phg(fg,mg,cg) = Prob(h|g);
  */
float phg(fg, mg, cg)
int fg, mg, cg;
{
    if ((fg == -1) | (mg == -1) | (cg == -1)) {
        printf("ERROR: One of the genotypes used to calculate prob(h_j|g)\n");
        printf("       is missing\n");
        exit(1);
    }
    return phgd1[9 * fg + 3 * mg + cg + 1];
}                               //END phg

/**   phgz: given the genotype of the father, the genotype of the mother,
    the genotype of a child, and the covariates, zz, associated with the
    child, phgz calculates p(cg | (fg,mg),zz). 
    nuiz = exp(\nu_i *zz).  For actual families it is being held 
    in findiv.
  */
float phgz(fg, mg, cg, nu1z, nu2z)
int fg, mg, cg;
float nu1z, nu2z;
{
    int i1;
    float tmp[3];
    tmp[0] = phg(fg, mg, 0);
    tmp[1] = phg(fg, mg, 1) * nu1z;
    tmp[2] = phg(fg, mg, 2) * nu2z;
    return tmp[cg] / (tmp[0] + tmp[1] + tmp[2]);
}                               //END phgz


/** fpi uses a matrix, ccovz, of covariate data --- actually only
                   findiv columns 2 and 3 for family i,
             a vector, xgeno, of possible genotypes for a family
                       (fg, mg, cg1,....,cgn)
                        where fg and/or mg may be missing (if mg is
                        missing then fg is missing also.  Code for
                        missing is -1),
             the number, nninfam, of lines of covariate data,
             the number, xnpgeno, of possible parental genotypes given
                        xgeno,
             a matrix xpgenos of possible parental genotypes (fg1, mg1)
                                                             .......... 
                                                             (fgn, mgn). 
       For each row of xpgenos it calculates 
              Prob((xpgenos[1],xpgenos[2])| xgeno, ppars, ccovz). 
              These are the "weights" needed when one or both of the 
              parental genotypes are missing (or 1 if none are missing.

        It stores (ppgenos[1], ppgenos[2],
               Prob((ppgenos[1], ppgenos[2])| ffamgeno, ppars, ccovz))
                  in row i of the matrix ffgiout.
   */

void fpi(ffdpm, xfamgeno, xpgenos, xnpgeno, ffpgiout)
struct famdat *ffdpm;
float **ffpgiout;
int *xfamgeno;
int **xpgenos;
int xnpgeno;
{
    int *ffamgeno;
    int nninfam;
    int nnpgeno;
    int **ppgenos;
    float **ffindiv;
    float tmp1, tmp2, tmp3;
    int i1, i2, i3;

    ffamgeno = xfamgeno;
    nninfam = (*ffdpm).ninfam;
    nnpgeno = xnpgeno;
    ppgenos = xpgenos;
    ffindiv = (*ffdpm).findiv;

    tmp2 = 0;
    for (i1 = 1; i1 <= nnpgeno; i1++) {
        tmp1 = 1;
        for (i2 = 3; i2 <= nninfam; i2++) {
/*
 printf("TOHERE1 %ld %ld\n", ppgenos[i1][1],ppgenos[i1][2]);
 printf("TOHERE2 %ld\n", ffamgeno[i2]);
 printf("TOHERE3 %lf %lf\n", ffindiv[i2][2],ffindiv[i2][3]);
*/
            tmp1 *=
                phgz(ppgenos[i1][1], ppgenos[i1][2], ffamgeno[i2],
                     ffindiv[i2][2], ffindiv[i2][3]);

/*
if((*ffdpm).famid == 19 && ckmkuih == 1){
printf("\n");
printf("TOHEREtst  %lf %lf %lf\n",phgz(1,0,0,1.5,1.0),phgz(1,0,1,1.5,1.0),
                   phgz(1,0,2,1.45,1.0));
printf("TOHEREtst1  %lf %lf %lf\n",phg(0,0,0),phg(0,0,1), phg(0,0,2));
printf("TOHEREffgeno  %ld  %ld  %ld \n",ppgenos[i1][1],ppgenos[i1][2],
                                      ffamgeno[3]);
printf("TOHERE tmp1 %lf %lf %lf\n",tmp1, ffindiv[i2][2],ffindiv[i2][3]);
}
*/

        }
        tmp1 *= gammma[ppgenos[i1][1]] * gammma[ppgenos[i1][2]];
        ffpgiout[i1][1] = ppgenos[i1][1];
        ffpgiout[i1][2] = ppgenos[i1][2];
        ffpgiout[i1][3] = tmp1;
        tmp2 += tmp1;
    }

    for (i1 = 1; i1 <= nnpgeno; i1++) {
        ffpgiout[i1][3] /= tmp2;
    }

}                               //END fpi


float mkmuijg(g, nu1z, nu2z)
int *g;
float nu1z, nu2z;
{
    float pp0, pp1, pp2;
    pp0 = phgz(g[1], g[2], 0, nu1z, nu2z);
    pp1 = phgz(g[1], g[2], 1, nu1z, nu2z);
    pp2 = phgz(g[1], g[2], 2, nu1z, nu2z);

/*
if(ckmkuih == 1 && whichfam == 19){
printf("TOHEREinmu0 g1=%ld g2=%ld \n",
                     g[1],g[2]);
printf("TOHEREinmu pp0=%lf pp1=%lf pp2=%lf n1=%lf n2=%lf\n",
            pp0,pp1,pp2,nu1z,nu2z);
}
*/

    return cch[0] * pp0 + cch[1] * pp1 + cch[2] * pp2;
}                               //END mkmuijg

float mkmuu1ijg(g, nu1z, nu2z)
int *g;
float nu1z, nu2z;
{
    float pp0, pp1, pp2;
    pp0 = phgz(g[1], g[2], 0, nu1z, nu2z);
    pp1 = phgz(g[1], g[2], 1, nu1z, nu2z);
    pp2 = phgz(g[1], g[2], 2, nu1z, nu2z);

/*
if(ckmkuih == 1 && whichfam == 19){
printf("TOHEREinmu0 g1=%ld g2=%ld \n",
                     g[1],g[2]);
printf("TOHEREinmu pp0=%lf pp1=%lf pp2=%lf n1=%lf n2=%lf\n",
            pp0,pp1,pp2,nu1z,nu2z);
}
*/

    return cchu1[0] * pp0 + cchu1[1] * pp1 + cchu1[2] * pp2;
}                               //END mkmuu1ijg


void mkfindiv(ppars, ffdpm)
struct famdat *ffdpm;
float *ppars;
{
    float tmp1, tmp2, tmp3;
    int i, i1, j, itmp1;
//This pars is here to be sure that the global
//    pars is insulated from this function.
    float pars[100];

//JH FILL pars from ogpars<-->ppars  in mkfindiv

if(cmodeltype == 0){
//UNCONSTRAINED
     for(i = 1; i <= npars-2; i++){
       pars[i] = ppars[i];
     } 
}
else if(cmodeltype == 1){
//RECESSIVE
     for(i = 1; i <= npars-2; i++){
       pars[i] = 0;
       pars[i + (npars-2)] = ppars[i];
     } 
}
else if(cmodeltype == 2){
//DOMINANT
     for(i = 1; i <= npars-2; i++){
       pars[i] = ppars[i];
       pars[i + (npars-2)] = ppars[i];
     } 
}
else if(cmodeltype == 3){
//ADDITIVE
     for(i = 1; i <= npars-2; i++){
       pars[i] = ppars[i];
       pars[i + (npars-2)] = 2 * ppars[i];
     } 
}

    pars[2 * ncov + 2] = ppars[npars];
    pars[2 * ncov + 1] = ppars[npars - 1];

    for (j = 1; j <= (*fdp).ninfam; j++) {
        tmp1 = 0;
        tmp2 = 0;
        tmp3 = 0;
        for (i1 = 1; i1 <= ncov; i1++) {
            tmp1 += pars[i1] * (*ffdpm).fcovdat[j][i1];
            tmp2 += pars[ncov + i1] * (*ffdpm).fcovdat[j][i1];
            tmp3 += deltas[i1] * (*ffdpm).fcovdat[j][i1];
        }
//(*ffdpm).findiv[j][1] = 1.0;
        (*ffdpm).findiv[j][2] = exp(tmp1);
        (*ffdpm).findiv[j][3] = exp(tmp2);

/*
//Commented Out because, in this program version, findiv[j][4] is
//  given by the wweights.dat file. See above where it is read in.
        if ((*ffdpm).phenodat[j] != 0) {
            tmp3 = exp(rho0 + tmp3);
            tmp3 = tmp3 / (1 + tmp3);
            itmp1 = (*ffdpm).phenodat[j] - 1;
            (*ffdpm).findiv[j][4] = itmp1 * (1 - tmp3) - (1 - itmp1) * tmp3;
        } else {
            (*ffdpm).findiv[j][4] = 0;
        }
*/
        if ((*ffdpm).phenodat[j] == 0) {
            (*ffdpm).findiv[j][4] = 0;
        }
//So findiv[j][4] is the individual's "weight". cf.(5) TDT memo.
    }
}                               //END mkfindiv



float loglik(ppars)
float *ppars;
{
    int i1, i2, i3, i4, i;
    float tmp1, tmp2, tmpm, tmps;
    float loglkk = 0;
    float pparg1, pparg2;
    float tmpp[3];

    pparg1 = (atan(ppars[npars - 1]) + 3.14159 / 2.0) / 3.14159;
    pparg2 = 1 - pparg1;
    pparg2 = pparg2 * (atan(ppars[npars]) + 3.14159 / 2.0) / 3.14159;

    for (i1 = 1; i1 <= nfam; i1++) {
        fdp = &ffdp[i1];
        mkfindiv(ppars, fdp);
        tmps = 0;
        for (i2 = 1; i2 <= (*fdp).npgeno; i2++) {
            tmpm = 1;
            for (i3 = 3; i3 <= (*fdp).ninfam; i3++) {
                tmpm *=
                    phgz((*fdp).pgenos[i2][1], (*fdp).pgenos[i2][2],
                         (*fdp).famgeno[i3], (*fdp).findiv[i3][2],
                         (*fdp).findiv[i3][3]);
            }
            tmpp[0] = 1.0 - pparg1 - pparg2;
            tmpp[1] = pparg1;
            tmpp[2] = pparg2;

            tmps +=
                tmpp[(*fdp).pgenos[i2][1]] * tmpp[(*fdp).pgenos[i2][2]] *
                tmpm;
        }
        loglkk += log(tmps);
    }                           //END nfam
    return (-loglkk);
}                               //END loglik


void mkuih(ffdpm, xgeno)
struct famdat *ffdpm;
int *xgeno;
{
    whichfam = (*ffdpm).famid;
    float mu = 0;
    float muu1 = 0;
    float a1, a2;
    int i, i1, i2, j;

    for (i1 = 1; i1 <= npars + 1; i1++) {
        uih[i1] = 0;
    }

    for (j = 3; j <= (*ffdpm).ninfam; j++) {

        muu1 = mkmuu1ijg(xgeno, (*ffdpm).findiv[j][2], (*ffdpm).findiv[j][3]);
        mu = mkmuijg(xgeno, (*ffdpm).findiv[j][2], (*ffdpm).findiv[j][3]);

// Code for uih[1] for both the constrained and unconstrained cases.
        uih[1] += (*ffdpm).findiv[j][4] * (cchu1[xgeno[j]] - muu1);

/*
if(ckmkuih == 1 && (*ffdpm).famid == 19){
printf("TOHERE uihh findiv4=%lf mu=%lf cchxgeno=%ld  xgeno=%ld uih=%lf\n",
                    (*ffdpm).findiv[j][4],mu,cch[xgeno[j]],xgeno[j],uih[1]);
printf("TOHEREuihh1 xgen1=%ld %ld %ld findiv2=%lf %lf\n",
             xgeno[1],xgeno[2],xgeno[3],
         (*ffdpm).findiv[j][2], (*ffdpm).findiv[j][3]);
}
*/

//When nu's are constrainted,  change code appropriately between
//JH-UIH "=========================" to compute uih[2,...,npars-1].
//==============================================================
if(cmodeltype == 0){
// Code for uih[j], j> 1, unconstrained case when npars = 2*ncov + 2.
        a1 = a2 = 0;
        if (xgeno[j] == 1) {
            a1 = 1;
        } else if (xgeno[j] == 2) {
            a2 = 1;
        }
        a1 = a1 - phgz(xgeno[1], xgeno[2], 1,
                       (*ffdpm).findiv[j][2], (*ffdpm).findiv[j][3]);
        a2 = a2 - phgz(xgeno[1], xgeno[2], 2,
                       (*ffdpm).findiv[j][2], (*ffdpm).findiv[j][3]);
        for (i1 = 1; i1 <= ncov; i1++) {
            uih[1 + i1] += (*ffdpm).fcovdat[j][i1] * a1;
            uih[1 + i1 + ncov] += (*ffdpm).fcovdat[j][i1] * a2;
        }
  }
else if(cmodeltype > 0){
// Code for uih[j], with j > 1, constrained "dominant", recessive, 
//         or additive cases.   npars = ncov + 2.
        a1 = cch[xgeno[j]] - mu;
        for (i1 = 1; i1 <= ncov; i1++) {
            uih[1 + i1] += (*ffdpm).fcovdat[j][i1] * a1;
        }
}
//==============================================================
    }                           //END j (over children) loop
    i1 = 3 * xgeno[1] + xgeno[2] + 1;
    uih[npars] = -utilg[i1][1] / (1 - gammma[1] - gammma[2]) +
        utilg[i1][2] / gammma[1];
    uih[npars + 1] = -utilg[i1][1] / (1 - gammma[1] - gammma[2]) +
        utilg[i1][3] / gammma[2];
}                               //END mkuih 
